(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     16834,        496]
NotebookOptionsPosition[     14592,        425]
NotebookOutlinePosition[     15102,        445]
CellTagsIndexPosition[     15059,        442]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Example 2", "Subsubtitle",
 CellChangeTimes->{{3.561784903322611*^9, 3.561784914734438*^9}, {
  3.5855442449811163`*^9, 3.585544245709262*^9}, {3.585544364225713*^9, 
  3.585544364400597*^9}, {3.585544546786407*^9, 3.585544547049386*^9}}],

Cell[CellGroupData[{

Cell["The Fourier transform", "Title",
 CellChangeTimes->{{3.561784933316359*^9, 3.561784950590352*^9}, {
  3.5617882095607243`*^9, 3.561788217874277*^9}, {3.585537636231443*^9, 
  3.585537653026313*^9}, {3.585542528085225*^9, 3.58554253224314*^9}, {
  3.585544162354004*^9, 3.5855441725463533`*^9}, {3.5855445490096292`*^9, 
  3.585544552121216*^9}}],

Cell["\<\
This example demonstrates, that the Fourier transform can be used fortwo \
different kinds of data sets.\
\>", "Subsubtitle",
 CellChangeTimes->{{3.561788081362376*^9, 3.56178816956818*^9}, {
  3.5855441822319193`*^9, 3.5855442385494347`*^9}, {3.5855445841757603`*^9, 
  3.5855446199258966`*^9}}],

Cell["\<\
Author: \t\tRonny Bergmann
Created: \t\t15.08.2013
Last Changed: \t15.08.2013\
\>", "Program",
 CellChangeTimes->{{3.585540461223832*^9, 3.585540470991144*^9}, 
   3.5855435520430107`*^9}],

Cell[CellGroupData[{

Cell["License", "Subsubsection",
 CellChangeTimes->{{3.585537355831193*^9, 3.585537363963532*^9}}],

Cell["\<\
    This file is part of MPAWL.
  
      MPAWL is free software : you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
      MPAWL is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
      You should have received a copy of the GNU General Public License
    along with the MPAWL. If not, see <http://www.gnu.org/licenses/>.\
\>", "Program",
 CellChangeTimes->{{3.585537405416844*^9, 3.585537461767535*^9}, {
  3.585537502312166*^9, 3.585537513376918*^9}, {3.585537694042321*^9, 
  3.5855377067755623`*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Loading the Library", "Subsection",
 CellChangeTimes->{{3.561785031749501*^9, 3.561785066774623*^9}, {
  3.585537743617692*^9, 3.585537746062161*^9}}],

Cell[TextData[{
 "The MPAWL is located in the parent directory (see ",
 StyleBox["MPAWL.m", "Program"],
 ") in order to load the library, we add its path to ",
 StyleBox["$Path", "Code"],
 "."
}], "Text",
 CellChangeTimes->{{3.585537766270393*^9, 3.585537868296885*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$Path", " ", "=", " ", 
   RowBox[{"Join", "[", 
    RowBox[{"$Path", ",", 
     RowBox[{"{", 
      RowBox[{"ParentDirectory", "[", 
       RowBox[{"NotebookDirectory", "[", "]"}], "]"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SetDirectory", "[", 
    RowBox[{"NotebookDirectory", "[", "]"}], "]"}], ";"}], 
  RowBox[{"(*", 
   RowBox[{"Set", " ", "to", " ", "actual", " ", "directory"}], 
   "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"Needs", "[", "\"\<MPAWL`\>\"", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.561784454440884*^9, 3.561784464446355*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The data set as a matrix", "Subsection",
 CellChangeTimes->{{3.585542611369248*^9, 3.585542622864993*^9}, {
  3.5855446303364973`*^9, 3.58554463850933*^9}}],

Cell["\<\
Let' s look at a matrix having more than one cycle (in contrast to the matrix \
from Example 1). For\
\>", "Text",
 CellChangeTimes->{{3.5855446450184097`*^9, 3.585544679011527*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"mM", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"16", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "16"}], "}"}]}], "}"}]}], ";", " ", 
  RowBox[{"MatrixForm", "[", "mM", "]"}]}]], "Input",
 CellChangeTimes->{{3.585544674567384*^9, 3.585544691556964*^9}}],

Cell["We have", "Text",
 CellChangeTimes->{{3.585544693627241*^9, 3.585544694267748*^9}, 
   3.58554473393038*^9}],

Cell[BoxData[
 RowBox[{"patternDimension", "[", "mM", "]"}]], "Input",
 CellChangeTimes->{{3.5855446991123743`*^9, 3.5855447032337303`*^9}}],

Cell["and", "Text",
 CellChangeTimes->{{3.585544705244916*^9, 3.585544733930869*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    SubscriptBox["v", "1"], ",", 
    SubscriptBox["v", "2"]}], "}"}], " ", "=", " ", 
  RowBox[{"patternBasis", "[", "mM", "]"}]}]], "Input",
 CellChangeTimes->{{3.5855447104346848`*^9, 3.5855447153384657`*^9}, {
  3.5855455451293497`*^9, 3.585545562553149*^9}}],

Cell["\<\
where the elementary divisors are\
\>", "Text",
 CellChangeTimes->{{3.585544738849497*^9, 3.585544755392313*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    SubscriptBox["\[Epsilon]", "1"], ",", 
    SubscriptBox["\[Epsilon]", "2"]}], "}"}], " ", "=", " ", 
  RowBox[{"Diagonal", "[", 
   RowBox[{"IntegerSmithForm", "[", 
    RowBox[{"mM", ",", 
     RowBox[{"ExtendedForm", "\[Rule]", " ", "False"}]}], "]"}], 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.585544756954894*^9, 3.585544777682675*^9}, {
  3.585545608215355*^9, 3.585545615871068*^9}}],

Cell["and hence", "Text",
 CellChangeTimes->{{3.585545399365362*^9, 3.585545401975342*^9}, 
   3.585547353923274*^9}],

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{"pattern", "[", 
   RowBox[{"getPatternNormalform", "[", "mM", "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.5855454028999968`*^9, 3.58554544744916*^9}}],

Cell[TextData[{
 "is the same as (where ",
 StyleBox["Flatten[...,1]", "Code"],
 " is used to get a vector of points instead of a matrix)"
}], "Text",
 CellChangeTimes->{{3.585545520776881*^9, 3.585545522254546*^9}, {
  3.5855457154554033`*^9, 3.5855457600712852`*^9}}],

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{"Flatten", "[", 
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"modM", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          SubscriptBox["k", "1"], "*", 
          SubscriptBox["v", "1"]}], "+", 
         RowBox[{
          SubscriptBox["k", "2"], 
          SubscriptBox["v", "2"]}]}], ",", 
        RowBox[{"IdentityMatrix", "[", "2", "]"}], ",", 
        RowBox[{"Target", "\[Rule]", " ", "\"\<Symmetric\>\""}]}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        SubscriptBox["k", "1"], ",", "0", ",", 
        RowBox[{
         SubscriptBox["\[Epsilon]", "1"], "-", "1"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        SubscriptBox["k", "2"], ",", "0", ",", 
        RowBox[{
         SubscriptBox["\[Epsilon]", "2"], "-", "1"}]}], "}"}]}], "]"}], ",", 
    "1"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.5855455271794567`*^9, 3.585545658174035*^9}, {
  3.585545696804769*^9, 3.585545703812462*^9}}],

Cell[BoxData[
 StyleBox[
  RowBox[{
   SuperscriptBox[
    RowBox[{
     RowBox[{
     "If", " ", "we", " ", "think", " ", "of", " ", "these", " ", "points", 
      " ", "as", " ", "sampling", " ", "points", " ", "and", " ", "scale", 
      " ", "them", " ", "onto", " ", "the", " ", 
      RowBox[{"torus", " ", "[", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}]}]}], ")"}], "2"], ",", " ", 
   RowBox[{
    RowBox[{"i", ".", "e", ".", " ", "by"}], " ", "multiplying", " ", "them", 
    " ", "with", " ", "2", "\[Pi]"}], ",", " ", 
   RowBox[{
    RowBox[{
    "we", " ", "obtain", " ", "a", " ", "matrix", " ", "of", " ", "dimension",
      " ", 
     SubscriptBox["\[Epsilon]", "1"], "\[Times]", 
     SubscriptBox["\[Epsilon]", "2"]}], " ", "=", " ", 
    RowBox[{
    "4", "\[Times]", "64", "  ", "of", " ", "sampling", " ", "values"}]}], 
   ",", " ", 
   RowBox[{"for", " ", "example"}]}],
  FontFamily->"Arial"]], "Text",
 CellChangeTimes->{{3.5855457861983633`*^9, 3.585545911169107*^9}, {
  3.585546956811541*^9, 3.585546976040462*^9}, {3.5855474821231117`*^9, 
  3.58554749548534*^9}}],

Cell[BoxData[
 RowBox[{"b", " ", "=", " ", 
  RowBox[{"Table", "[", " ", 
   RowBox[{
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["k", "1"], "==", " ", "0"}], ")"}], "&&", 
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["k", "2"], "==", " ", "0"}], ")"}]}], ",", "1", ",", 
      "0"}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{
      SubscriptBox["k", "1"], ",", "0", ",", 
      RowBox[{
       SubscriptBox["\[Epsilon]", "1"], "-", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      SubscriptBox["k", "2"], ",", "0", ",", 
      RowBox[{
       SubscriptBox["\[Epsilon]", "2"], "-", "1"}]}], "}"}]}], 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.585545958981542*^9, 3.5855460224548388`*^9}}],

Cell["We obtain its Fourier Transform with respect to mM by", "Text",
 CellChangeTimes->{{3.585546038925001*^9, 3.585546057943142*^9}}],

Cell[BoxData[
 RowBox[{"?", "FourierTransformTorus"}]], "Input",
 CellChangeTimes->{{3.585546102055772*^9, 3.58554610599687*^9}}],

Cell[BoxData[
 RowBox[{"hatb", " ", "=", " ", 
  RowBox[{"FourierTransformTorus", "[", 
   RowBox[{"mM", ",", "b"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.585546060684635*^9, 3.5855460920509033`*^9}}],

Cell["which can also be switched to exact computations", "Text",
 CellChangeTimes->{{3.585546143833268*^9, 3.5855461538978033`*^9}, 
   3.585547324134453*^9}],

Cell[BoxData[
 RowBox[{"hatb", " ", "=", " ", 
  RowBox[{"FourierTransformTorus", "[", 
   RowBox[{"mM", ",", "b", ",", " ", 
    RowBox[{"Compute", "\[Rule]", " ", "\"\<Exact\>\""}]}], "]"}]}]], "Input",\

 CellChangeTimes->{{3.5855461334750557`*^9, 3.585546137385312*^9}}],

Cell["\<\
Of course now, the values correspond to the same order used above with \
respect to the basis of the generating set\
\>", "Text",
 CellChangeTimes->{{3.585548248447962*^9, 3.585548301514982*^9}}],

Cell[BoxData[
 RowBox[{"generatingSetBasis", "[", 
  RowBox[{"Transpose", "[", "mM", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.585548281368471*^9, 3.585548291406822*^9}}],

Cell["Due to ", "Text",
 CellChangeTimes->{{3.5855462882349052`*^9, 3.585546288922772*^9}, 
   3.585547335023238*^9}],

Cell[BoxData[
 RowBox[{"Abs", "[", 
  RowBox[{"Det", "[", "mM", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.585546292152486*^9, 3.585546295107946*^9}}],

Cell["\<\
this is of course the unitary version of the Fourier transform and hence\
\>", "Text",
 CellChangeTimes->{{3.585546298892672*^9, 3.58554630992238*^9}, 
   3.5855469452957582`*^9}],

Cell[BoxData[
 RowBox[{"FourierTransformTorus", "[", 
  RowBox[{"mM", ",", "hatb", ",", " ", 
   RowBox[{"Compute", "\[Rule]", " ", "\"\<Exact\>\""}]}], "]"}]], "Input",
 CellChangeTimes->{{3.5855462751809998`*^9, 3.5855462776595287`*^9}, {
  3.5855463134180927`*^9, 3.5855463271460867`*^9}}],

Cell[TextData[{
 "But we could also reshape the data to be a vector by ordering the indices \
",
 StyleBox["{0, 0}, ..., {3, 63}", "Code"],
 " lexicographically, i.e. ",
 StyleBox["{0, 0}, {0, 1}, ..., {1, 0}, {1, 1}, ... {3, 62}, {3, 63}", "Code"],
 ". For details see [1]."
}], "Text",
 CellChangeTimes->{{3.585546392635366*^9, 3.5855464854895697`*^9}, {
  3.58557250025395*^9, 3.5855725205922327`*^9}}],

Cell[BoxData[
 RowBox[{"?", "reshapeData"}]], "Input",
 CellChangeTimes->{{3.585546371910028*^9, 3.585546377072661*^9}}],

Cell[BoxData[
 RowBox[{"b2", " ", "=", " ", 
  RowBox[{"reshapeData", "[", 
   RowBox[{"mM", ",", "b", ",", "False"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.585546360540142*^9, 3.585546370039955*^9}, {
  3.585546503114584*^9, 3.585546512603394*^9}, {3.585546607120615*^9, 
  3.585546607821887*^9}}],

Cell[BoxData[
 RowBox[{"hatb2", " ", "=", " ", 
  RowBox[{"FourierTransformTorus", "[", 
   RowBox[{"mM", ",", "b2"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.585546598722905*^9, 3.585546634046604*^9}, {
  3.585546793382532*^9, 3.585546801462391*^9}, {3.5855468349491587`*^9, 
  3.585546837094754*^9}}],

Cell[TextData[{
 "again using the above ordering of the indices but now with respect to the \
generating set of ",
 StyleBox["Transpose[mM]", "Code"],
 ". Of course the inverse transform yields again "
}], "Text",
 CellChangeTimes->{{3.5855468774691553`*^9, 3.5855468879943237`*^9}, 
   3.585546935657735*^9, {3.585548329606914*^9, 3.585548361813231*^9}}],

Cell[BoxData[
 RowBox[{"FourierTransformTorus", "[", 
  RowBox[{"mM", ",", "hatb2"}], "]"}]], "Input",
 CellChangeTimes->{{3.585546855411831*^9, 3.585546858466866*^9}, {
  3.585547239796562*^9, 3.5855473022363987`*^9}, {3.585547563705112*^9, 
  3.585547565469406*^9}}],

Cell["\<\
but of course this works if and only if the ordering of the elements in the \
vector is as just explained. Notice that while the input shape is analyzed \
automatically, i.e. whether its a vector or a matrix of data, the usual \
computation is always done numerically and the exact computation must be \
activated by the option Compute-> . Also the exact computation yields $Failed \
if any numeric number is given as does the numeric one if any element is not \
a number.\
\>", "Text",
 CellChangeTimes->{{3.5855468913053017`*^9, 3.585546929703498*^9}, {
  3.585547032323731*^9, 3.585547063915085*^9}, {3.5855471163200483`*^9, 
  3.5855472635861893`*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Literature", "Subsection",
 CellChangeTimes->{{3.585572592950934*^9, 3.585572600302207*^9}}],

Cell[TextData[{
 "[1] R.Bergmann, ",
 StyleBox["The fast Fourier transform and fast wavelet transform for patterns \
on the torus",
  FontSlant->"Italic"],
 ", Appl. Comp. Harmon. Anal. 35 (2013) 39\[Dash]51, doi: ",
 ButtonBox["10.1016/j.acha .2012 .07 .007",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://dx.doi.org/10.1016/j.acha.2012.07.007"], None},
  ButtonNote->"http://dx.doi.org/10.1016/j.acha.2012.07.007"],
 "."
}], "Text",
 CellChangeTimes->{{3.585572609227313*^9, 3.585572616009447*^9}, {
  3.585572717305231*^9, 3.5855727173078423`*^9}, {3.585573098697709*^9, 
  3.5855731192963257`*^9}}]
}, Open  ]]
}, Open  ]]
},
WindowSize->{1264, 714},
WindowMargins->{{87, Automatic}, {Automatic, 77}},
PrintingCopies->1,
PrintingPageRange->{1, Automatic},
PrivateNotebookOptions->{"VersionedStylesheet"->{"Default.nb"[8.] -> False}},
ShowSelection->True,
FrontEndVersion->"9.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (January 25, \
2013)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[557, 20, 244, 3, 29, "Subsubtitle"],
Cell[CellGroupData[{
Cell[826, 27, 351, 5, 92, "Title"],
Cell[1180, 34, 306, 6, 29, "Subsubtitle"],
Cell[1489, 42, 198, 6, 74, "Program"],
Cell[CellGroupData[{
Cell[1712, 52, 98, 1, 35, "Subsubsection"],
Cell[1813, 55, 872, 18, 228, "Program"]
}, Closed]],
Cell[CellGroupData[{
Cell[2722, 78, 156, 2, 36, "Subsection"],
Cell[2881, 82, 270, 7, 31, "Text"],
Cell[3154, 91, 635, 18, 63, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3826, 114, 162, 2, 44, "Subsection"],
Cell[3991, 118, 192, 4, 30, "Text"],
Cell[4186, 124, 349, 10, 28, "Input"],
Cell[4538, 136, 114, 2, 30, "Text"],
Cell[4655, 140, 140, 2, 28, "Input"],
Cell[4798, 144, 85, 1, 30, "Text"],
Cell[4886, 147, 315, 8, 28, "Input"],
Cell[5204, 157, 123, 3, 30, "Text"],
Cell[5330, 162, 448, 12, 28, "Input"],
Cell[5781, 176, 117, 2, 30, "Text"],
Cell[5901, 180, 211, 4, 28, "Input"],
Cell[6115, 186, 269, 6, 30, "Text"],
Cell[6387, 194, 1016, 29, 28, "Input"],
Cell[7406, 225, 1114, 28, 54, "Text"],
Cell[8523, 255, 788, 25, 28, "Input"],
Cell[9314, 282, 135, 1, 30, "Text"],
Cell[9452, 285, 129, 2, 28, "Input"],
Cell[9584, 289, 203, 4, 28, "Input"],
Cell[9790, 295, 158, 2, 30, "Text"],
Cell[9951, 299, 274, 6, 28, "Input"],
Cell[10228, 307, 205, 4, 30, "Text"],
Cell[10436, 313, 174, 3, 28, "Input"],
Cell[10613, 318, 117, 2, 30, "Text"],
Cell[10733, 322, 153, 3, 28, "Input"],
Cell[10889, 327, 189, 4, 30, "Text"],
Cell[11081, 333, 292, 5, 28, "Input"],
Cell[11376, 340, 405, 9, 30, "Text"],
Cell[11784, 351, 120, 2, 28, "Input"],
Cell[11907, 355, 301, 6, 28, "Input"],
Cell[12211, 363, 303, 6, 28, "Input"],
Cell[12517, 371, 355, 7, 30, "Text"],
Cell[12875, 380, 268, 5, 28, "Input"],
Cell[13146, 387, 666, 11, 68, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[13849, 403, 98, 1, 44, "Subsection"],
Cell[13950, 406, 614, 15, 30, "Text"]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
