(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     27730,        781]
NotebookOptionsPosition[     25113,        697]
NotebookOutlinePosition[     25777,        723]
CellTagsIndexPosition[     25692,        718]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Example 4", "Subsubtitle",
 CellChangeTimes->{{3.561784903322611*^9, 3.561784914734438*^9}, {
  3.5855442449811163`*^9, 3.585544245709262*^9}, {3.585544364225713*^9, 
  3.585544364400597*^9}, {3.585544546786407*^9, 3.585544547049386*^9}, {
  3.585575431155713*^9, 3.585575431493561*^9}}],

Cell[CellGroupData[{

Cell["Sampling and Interpolation", "Title",
 CellChangeTimes->{{3.561784933316359*^9, 3.561784950590352*^9}, {
  3.5617882095607243`*^9, 3.561788217874277*^9}, {3.585537636231443*^9, 
  3.585537653026313*^9}, {3.585542528085225*^9, 3.58554253224314*^9}, {
  3.585544162354004*^9, 3.5855441725463533`*^9}, {3.5855445490096292`*^9, 
  3.585544552121216*^9}, {3.585575420661229*^9, 3.58557542659937*^9}}],

Cell["\<\
Taking a linear function and Box Spline, sampling it and projecting into a \
certain space of translates.\
\>", "Subsubtitle",
 CellChangeTimes->{{3.561788081362376*^9, 3.56178816956818*^9}, {
  3.5855441822319193`*^9, 3.5855442385494347`*^9}, {3.5855445841757603`*^9, 
  3.5855446199258966`*^9}, {3.5855754351385202`*^9, 3.5855754749521427`*^9}}],

Cell["\<\
 Author: \t\tRonny Bergmann
Created: \t\t15.08.2013
Last Changed: \t15.08.2013\
\>", "Program",
 CellChangeTimes->{{3.585540461223832*^9, 3.585540470991144*^9}, 
   3.5855435520430107`*^9, 3.585575663357239*^9}],

Cell[CellGroupData[{

Cell["License", "Subsubsection",
 CellChangeTimes->{{3.585537355831193*^9, 3.585537363963532*^9}}],

Cell["\<\
    This file is part of MPAWL.
  
      MPAWL is free software : you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
      MPAWL is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
      You should have received a copy of the GNU General Public License
    along with the MPAWL. If not, see <http://www.gnu.org/licenses/>.\
\>", "Program",
 CellChangeTimes->{{3.585537405416844*^9, 3.585537461767535*^9}, {
  3.585537502312166*^9, 3.585537513376918*^9}, {3.585537694042321*^9, 
  3.5855377067755623`*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Loading the Library", "Subsection",
 CellChangeTimes->{{3.561785031749501*^9, 3.561785066774623*^9}, {
  3.585537743617692*^9, 3.585537746062161*^9}}],

Cell[TextData[{
 "The MPAWL is located in the parent directory (see ",
 StyleBox["MPAWL.m", "Program"],
 ") in order to load the library, we add its path to ",
 StyleBox["$Path", "Code"],
 "."
}], "Text",
 CellChangeTimes->{{3.585537766270393*^9, 3.585537868296885*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$Path", " ", "=", " ", 
   RowBox[{"Join", "[", 
    RowBox[{"$Path", ",", 
     RowBox[{"{", 
      RowBox[{"ParentDirectory", "[", 
       RowBox[{"NotebookDirectory", "[", "]"}], "]"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SetDirectory", "[", 
    RowBox[{"NotebookDirectory", "[", "]"}], "]"}], ";"}], 
  RowBox[{"(*", 
   RowBox[{"Set", " ", "to", " ", "actual", " ", "directory"}], 
   "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"Needs", "[", "\"\<MPAWL`\>\"", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.561784454440884*^9, 3.561784464446355*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Sampling a (centered) Box spline\
\>", "Subsection",
 CellChangeTimes->{{3.5855764514383183`*^9, 3.585576465704108*^9}, {
  3.585576536249794*^9, 3.5855765383979177`*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"m\[CapitalXi]", " ", "=", " ", 
   RowBox[{"\[Pi]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "1"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "/", "8"}], ",", "0"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"1", "/", "8"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"1", "/", "8"}], ",", 
        RowBox[{
         RowBox[{"-", "1"}], "/", "8"}]}], "}"}]}], "}"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ct", " ", "=", " ", 
   RowBox[{"Sum", "[", 
    RowBox[{
     RowBox[{"x", "/", "2"}], ",", "\[NonBreakingSpace]", 
     RowBox[{"{", 
      RowBox[{"x", ",", "m\[CapitalXi]"}], "}"}]}], "]"}]}], ";"}]}], "Input",\

 CellChangeTimes->{
  3.5039186842923098`*^9, {3.5039187219488373`*^9, 3.503918833562999*^9}, {
   3.5039189173584013`*^9, 3.503918935647999*^9}, {3.503919017060658*^9, 
   3.503919030284737*^9}, {3.5039191028630123`*^9, 3.5039191325023813`*^9}, {
   3.5039191712192087`*^9, 3.5039192004170103`*^9}, {3.503919251342844*^9, 
   3.503919281797329*^9}, {3.5039193785155687`*^9, 3.503919379091158*^9}, {
   3.503919456394064*^9, 3.5039194573123827`*^9}, {3.503919537226652*^9, 
   3.503919542612624*^9}, {3.503919773308127*^9, 3.503919776442552*^9}, {
   3.503919850548209*^9, 3.503919877861145*^9}, {3.50392003670537*^9, 
   3.5039200460915194`*^9}, {3.50392027305923*^9, 3.503920278826462*^9}, {
   3.5044996167151814`*^9, 3.504499619939143*^9}, {3.504502393573703*^9, 
   3.504502399628572*^9}, {3.5085845896176157`*^9, 3.508584589791716*^9}, {
   3.516687984502013*^9, 3.5166879854347363`*^9}, {3.518348753320738*^9, 
   3.5183487664700127`*^9}, {3.5604144238617373`*^9, 3.560414433533194*^9}, 
   3.560414567245927*^9, {3.565068965938684*^9, 3.5650689671198072`*^9}, 
   3.5650809101293497`*^9, 3.565243658529274*^9, {3.56978458719232*^9, 
   3.56978459391978*^9}, {3.5855765185881567`*^9, 3.585576587090193*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"f1", "[", "X_", "]"}], " ", ":=", 
  RowBox[{"evaluateBoxSpline", "[", 
   RowBox[{
    RowBox[{"N", "[", "m\[CapitalXi]", "]"}], ",", 
    RowBox[{"N", "[", 
     RowBox[{"X", "+", " ", "ct"}], "]"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.516601421145957*^9, 3.516601438277596*^9}, {
   3.516602124807948*^9, 3.5166021351090918`*^9}, 3.516602200594761*^9, {
   3.560414414088305*^9, 3.5604144405787497`*^9}, {3.5604147606044483`*^9, 
   3.5604147623731937`*^9}, {3.5604254386524553`*^9, 3.560425462465803*^9}, {
   3.560426946426818*^9, 3.5604269471452217`*^9}, 3.5604325087796507`*^9, {
   3.560435015969927*^9, 3.560435017871881*^9}, {3.560441529033148*^9, 
   3.5604415332167473`*^9}, {3.560486498876748*^9, 3.5604865034438763`*^9}, {
   3.571815990700861*^9, 3.571815999211589*^9}, {3.585576528420443*^9, 
   3.5855765296689672`*^9}, {3.585576574219466*^9, 3.585576605539569*^9}, 
   3.585576702515284*^9, {3.585577879122882*^9, 3.585577892060296*^9}}],

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{
   RowBox[{"f1", "[", 
    RowBox[{"{", 
     RowBox[{
      SubscriptBox["x", "1"], ",", 
      SubscriptBox["x", "2"]}], "}"}], "]"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "1"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "2"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"Ticks", " ", "\[Rule]", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", 
        RowBox[{"{", 
         RowBox[{"\[Pi]", ",", "\"\<\>\""}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "\[Pi]"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"1", "/", "2"}], ",", "1"}], "}"}]}], "}"}]}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.564900460723063*^9, 3.564900460988924*^9}, {
   3.564908519137034*^9, 3.564908522040477*^9}, 3.564909564769824*^9, {
   3.56509652158829*^9, 3.565096522060741*^9}, {3.5651548039901743`*^9, 
   3.565154805188757*^9}, {3.5651756675203657`*^9, 3.56517566839789*^9}, {
   3.565236808041092*^9, 3.565236809007689*^9}, {3.5652395279322033`*^9, 
   3.565239533363698*^9}, {3.5652398734752817`*^9, 3.5652398865700893`*^9}, {
   3.5652436635929813`*^9, 3.565243666135091*^9}, {3.56524949060079*^9, 
   3.565249492760524*^9}, 3.5652495671570883`*^9, 3.565252093669478*^9, {
   3.5652549615165987`*^9, 3.565254961966716*^9}, {3.571802444344022*^9, 
   3.57180245157767*^9}, 3.575049003207944*^9, 3.583728027997389*^9, {
   3.583802686179942*^9, 3.5838026880689077`*^9}, {3.583806142232151*^9, 
   3.583806147001336*^9}, {3.5839000010261297`*^9, 3.5839001409880533`*^9}, {
   3.5839002610273523`*^9, 3.583900261420002*^9}, 3.583901611416958*^9, 
   3.5839016527819147`*^9, {3.5839023001667757`*^9, 3.583902326694503*^9}, {
   3.5855765500346518`*^9, 3.5855765553878317`*^9}, 3.585576615585816*^9, {
   3.5855766526885233`*^9, 3.5855766673133497`*^9}, {3.585576712809093*^9, 
   3.585576754884853*^9}}],

Cell["\<\
We will first use sampling, that corresponds to an quadratic image of the Box \
spline\
\>", "Text",
 CellChangeTimes->{{3.585578209002482*^9, 3.5855782450338593`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"mM", " ", "=", " ", 
   RowBox[{"64", " ", 
    RowBox[{"IdentityMatrix", "[", "2", "]"}]}]}], ";", " ", 
  RowBox[{"MatrixForm", "[", "mM", "]"}]}]], "Input",
 CellChangeTimes->{{3.585578037685845*^9, 3.585578091127769*^9}}],

Cell["\<\
Sampling the function using Sample including the possibility to save the \
samples to a file and load them from there, if that the samples can be \
loaded. The Debug can also provide times. Due to the long time, at least for \
Box splines, the samples are provided.\
\>", "Text",
 CellChangeTimes->{{3.585578393087749*^9, 3.58557850387066*^9}, {
  3.5855810330814943`*^9, 3.5855810488869963`*^9}}],

Cell[BoxData[
 RowBox[{"?", "sampleFunction"}]], "Input",
 CellChangeTimes->{{3.585578683408268*^9, 3.585578685480048*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"data", " ", "=", " ", 
   RowBox[{"sampleFunction", "[", 
    RowBox[{"mM", ",", "f1", ",", " ", 
     RowBox[{"Debug", "\[Rule]", " ", "\"\<Text&Time\>\""}], ",", " ", 
     RowBox[{"File", "\[Rule]", " ", "\"\<example4/sampling-f1.dat\>\""}]}], 
    "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.5855802096025343`*^9, 3.58558021550848*^9}}],

Cell[TextData[{
 "Using the Dirichlet - Kernel, we can also obtain the Bracket Sums at the \
same time by using the Option ",
 StyleBox["BracketSums", "Input"],
 "."
}], "Text",
 CellChangeTimes->{{3.585579345650447*^9, 3.585579418734551*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "DirichletKernel"}]], "Input",
 CellChangeTimes->{{3.5855787139624853`*^9, 3.5855787178997993`*^9}, {
   3.585579309238418*^9, 3.585579309935679*^9}, 3.5855805301162767`*^9}],

Cell[BoxData[
 StyleBox["\<\"DirichletKernel[mM]\\n\\nprovides a dirichlet kernel, which is \
a special case of the de la Vall\[EAcute]e Poussin mean,\\nwhere g=0, hence \
pyramidalFunction[d,0] is used. Here, the same options as for\\nthe de la \
Vallee Poussin mean apply.\"\>", "MSG"]], "Print", "PrintUsage",
 GeneratedCell->False,
 CellAutoOverwrite->False,
 CellChangeTimes->{3.585579435920492*^9, 3.585580530117189*^9},
 CellTags->"Info3585586635-4340400"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"{", 
    RowBox[{"ckDM", ",", "DMBS"}], "}"}], " ", "=", " ", 
   RowBox[{"DirichletKernel", "[", 
    RowBox[{"mM", ",", 
     RowBox[{"BracketSums", "\[Rule]", " ", "True"}]}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.585578697743208*^9, 3.5855787032562227`*^9}, {
  3.585579319892686*^9, 3.58557933792597*^9}}],

Cell["\<\
These coefficients are  again symmetric, as in Example 3, and hence the \
origin is in the center of the (in all dimensions) odd dimension lengths.\
\>", "Text",
 CellChangeTimes->{{3.5855797273254128`*^9, 3.585579778590424*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"origin", " ", "=", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Dimensions", "[", "ckDM", "]"}], "+", "1"}], ")"}], "/", 
    "2"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.5855797073385277`*^9, 3.585579725276128*^9}, {
  3.585579804172586*^9, 3.5855798055783873`*^9}, {3.58557985047777*^9, 
  3.585579851872903*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"max", " ", "=", " ", 
   RowBox[{"origin", "-", "1"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.5855798591183147`*^9, 3.585579861784153*^9}}],

Cell[TextData[{
 "Then, we can perform a change of basis, which is just a multiplication in \
the discrete Fourier domain. This can be done using ",
 StyleBox["changeBasis[]", "Code"],
 ", see e.g. Section 2.1 in [1]."
}], "Text",
 CellChangeTimes->{{3.58558068191229*^9, 3.585580740523641*^9}, 
   3.585581192168778*^9, {3.585581459727477*^9, 3.585581474161235*^9}}],

Cell[BoxData[
 RowBox[{"?", "changeBasis"}]], "Input",
 CellChangeTimes->{{3.585580747825569*^9, 3.585580749759017*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"coeffs", " ", "=", " ", 
   RowBox[{"changeBasis", "[", 
    RowBox[{"mM", ",", "data", ",", " ", "DMBS", ",", 
     RowBox[{"Debug", "\[Rule]", " ", "\"\<Text&Time\>\""}], ",", 
     RowBox[{"Input", "\[Rule]", " ", "\"\<Time\>\""}]}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.585579479549769*^9, 3.58557949623709*^9}, 
   3.58558028807162*^9, 3.585580372851769*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"ckf1a", " ", "=", " ", 
   RowBox[{"getFourierFromSpace", "[", 
    RowBox[{"coeffs", ",", "ckDM", ",", "origin", ",", "mM"}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.585579603870535*^9, 3.585579607589077*^9}, {
   3.585579660071509*^9, 3.5855797042249804`*^9}, {3.585579814747507*^9, 
   3.585579816536366*^9}, 3.58558034985634*^9, {3.5855803800957527`*^9, 
   3.585580384213269*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f1approx", "[", "x_", "]"}], ":=", " ", 
   RowBox[{"Sum", "[", " ", 
    RowBox[{
     RowBox[{
      RowBox[{"ckf1a", "[", 
       RowBox[{"[", 
        RowBox[{"Sequence", "@@", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"]}], "}"}], "+", "origin"}], ")"}]}], "]"}],
        "]"}], 
      RowBox[{"Exp", "[", 
       RowBox[{"I", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           SubscriptBox["k", "1"], ",", 
           SubscriptBox["k", "2"]}], "}"}], ".", "x"}]}], "]"}]}], " ", ",", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["k", "1"], ",", 
       RowBox[{"-", 
        RowBox[{"max", "[", 
         RowBox[{"[", "1", "]"}], "]"}]}], ",", 
       RowBox[{"max", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["k", "2"], ",", 
       RowBox[{"-", 
        RowBox[{"max", "[", 
         RowBox[{"[", "2", "]"}], "]"}]}], ",", 
       RowBox[{"max", "[", 
        RowBox[{"[", "2", "]"}], "]"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.585579596352151*^9, 3.585579602684441*^9}, {
  3.5855796390811567`*^9, 3.585579643478169*^9}, {3.585579822668603*^9, 
  3.5855798259283123`*^9}, {3.585580415085321*^9, 3.585580418290534*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"f1aTerm", " ", "=", " ", 
   RowBox[{"Simplify", "[", 
    RowBox[{"f1approx", "[", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["x", "1"], ",", 
       SubscriptBox["x", "2"]}], "}"}], "]"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.5855798850732813`*^9, 3.585579905242916*^9}, {
  3.5855799456427813`*^9, 3.585579967489579*^9}, {3.585580057951703*^9, 
  3.58558006537677*^9}, {3.585580394068396*^9, 3.5855804001096907`*^9}}],

Cell["\<\
We use Chop to exclude the small imaginary parts, that occur during \
computations\
\>", "Text",
 CellChangeTimes->{{3.585580097585204*^9, 3.5855801442868156`*^9}, {
  3.585581124106193*^9, 3.5855811469650173`*^9}}],

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{
   RowBox[{"Chop", "[", "f1aTerm", "]"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "1"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "2"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"Ticks", " ", "\[Rule]", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", 
        RowBox[{"{", 
         RowBox[{"\[Pi]", ",", "\"\<\>\""}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "\[Pi]"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"1", "/", "2"}], ",", "1"}], "}"}]}], "}"}]}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.585579919418797*^9, 3.5855799296785097`*^9}, {
  3.58557998753788*^9, 3.585579991013768*^9}, {3.585580443411792*^9, 
  3.585580443625519*^9}, {3.5855811008414707`*^9, 3.5855811157141743`*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Sampling a linear function", "Subsection",
 CellChangeTimes->{{3.5855764514383183`*^9, 3.585576465704108*^9}, {
  3.585576536249794*^9, 3.5855765383979177`*^9}, {3.585580234610149*^9, 
  3.585580242935004*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"f2", "[", 
   RowBox[{"{", 
    RowBox[{"x_", ",", "y_"}], "}"}], "]"}], " ", ":=", " ", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"x", "+", "y"}], ")"}], "/", 
   RowBox[{"(", 
    RowBox[{"2", "\[Pi]"}], ")"}]}]}]], "Input",
 CellChangeTimes->{{3.5855790939248543`*^9, 3.585579213838605*^9}}],

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{
   RowBox[{"f2", "[", 
    RowBox[{"{", 
     RowBox[{
      SubscriptBox["x", "1"], ",", 
      SubscriptBox["x", "2"]}], "}"}], "]"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "1"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "2"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"Ticks", " ", "\[Rule]", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", 
        RowBox[{"{", 
         RowBox[{"\[Pi]", ",", "\"\<\>\""}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "\[Pi]"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "1"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"1", "/", "2"}]}], "}"}]}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.585579148858923*^9, 3.5855791495175056`*^9}, {
  3.585579225460827*^9, 3.585579228520948*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"data2", " ", "=", " ", 
   RowBox[{"sampleFunction", "[", 
    RowBox[{"mM", ",", "f2", ",", " ", 
     RowBox[{"Debug", "\[Rule]", " ", "\"\<Text&Time\>\""}], ",", " ", 
     RowBox[{"File", "\[Rule]", " ", "\"\<example4/sampling-f2.dat\>\""}]}], 
    "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.585578289121254*^9, 3.585578320214141*^9}, {
   3.585578949406435*^9, 3.585578970230309*^9}, {3.585579279977706*^9, 
   3.5855793010598516`*^9}, 3.585580198734105*^9}],

Cell["\<\
Here, we can again use the Bracket sum from the first function to perform a \
change of basis.\
\>", "Text",
 CellChangeTimes->{{3.5855802948737593`*^9, 3.585580341083931*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"coeffs2", " ", "=", " ", 
   RowBox[{"changeBasis", "[", 
    RowBox[{"mM", ",", "data2", ",", " ", "DMBS", ",", 
     RowBox[{"Debug", "\[Rule]", " ", "\"\<Text&Time\>\""}], ",", 
     RowBox[{"Input", "\[Rule]", " ", "\"\<Time\>\""}]}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.585579479549769*^9, 3.58557949623709*^9}, 
   3.58558028807162*^9, 3.5855817164994383`*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"ckf2a", " ", "=", " ", 
   RowBox[{"getFourierFromSpace", "[", 
    RowBox[{"coeffs2", ",", "ckDM", ",", "origin", ",", "mM"}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.585579603870535*^9, 3.585579607589077*^9}, {
   3.585579660071509*^9, 3.5855797042249804`*^9}, {3.585579814747507*^9, 
   3.585579816536366*^9}, 3.58558034985634*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f2approx", "[", "x_", "]"}], ":=", " ", 
   RowBox[{"Sum", "[", " ", 
    RowBox[{
     RowBox[{
      RowBox[{"ckf2a", "[", 
       RowBox[{"[", 
        RowBox[{"Sequence", "@@", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"]}], "}"}], "+", "origin"}], ")"}]}], "]"}],
        "]"}], 
      RowBox[{"Exp", "[", 
       RowBox[{"I", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           SubscriptBox["k", "1"], ",", 
           SubscriptBox["k", "2"]}], "}"}], ".", "x"}]}], "]"}]}], " ", ",", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["k", "1"], ",", 
       RowBox[{"-", 
        RowBox[{"max", "[", 
         RowBox[{"[", "1", "]"}], "]"}]}], ",", 
       RowBox[{"max", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["k", "2"], ",", 
       RowBox[{"-", 
        RowBox[{"max", "[", 
         RowBox[{"[", "2", "]"}], "]"}]}], ",", 
       RowBox[{"max", "[", 
        RowBox[{"[", "2", "]"}], "]"}]}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.585579596352151*^9, 3.585579602684441*^9}, {
  3.5855796390811567`*^9, 3.585579643478169*^9}, {3.585579822668603*^9, 
  3.5855798259283123`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"f2aTerm", " ", "=", " ", 
   RowBox[{"Simplify", "[", 
    RowBox[{"f2approx", "[", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["x", "1"], ",", 
       SubscriptBox["x", "2"]}], "}"}], "]"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.5855798850732813`*^9, 3.585579905242916*^9}, {
  3.5855799456427813`*^9, 3.585579967489579*^9}, {3.585580057951703*^9, 
  3.58558006537677*^9}}],

Cell["\<\
Here, we see that the discontinuities at the boundary are approximated by a \
finite Fourier series and hence continuous.\
\>", "Text",
 CellChangeTimes->{{3.585580097585204*^9, 3.5855801442868156`*^9}, {
  3.585580586444771*^9, 3.585580609201878*^9}, {3.5855818597797127`*^9, 
  3.5855818861377974`*^9}}],

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{"f2aTerm", ",", " ", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "1"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["x", "2"], ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}], ",", 
   RowBox[{"Ticks", " ", "\[Rule]", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", 
        RowBox[{"{", 
         RowBox[{"\[Pi]", ",", "\"\<\>\""}], "}"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "\[Pi]"}], ",", 
        RowBox[{
         RowBox[{"-", "\[Pi]"}], "/", "2"}], ",", "0", ",", 
        RowBox[{"\[Pi]", "/", "2"}], ",", "\[Pi]"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"1", "/", "2"}], ",", "1"}], "}"}]}], "}"}]}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.585579919418797*^9, 3.5855799296785097`*^9}, {
  3.58557998753788*^9, 3.585579991013768*^9}, {3.5855818484429903`*^9, 
  3.585581850768919*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Literature", "Subsection",
 CellChangeTimes->{{3.585581389389732*^9, 3.58558139583228*^9}}],

Cell[TextData[{
 "[1] R.Bergmann, ",
 StyleBox["Translationsinvariante R\[ADoubleDot]ume multivariater anisotroper \
Funktionen auf dem Torus",
  FontSlant->"Italic"],
 ", Ph.D.thesis, University of L\[UDoubleDot]beck, 2013 (german)."
}], "Text",
 CellChangeTimes->{{3.585581405515045*^9, 3.585581449400193*^9}, {
  3.585581489585037*^9, 3.5855814899337883`*^9}}]
}, Open  ]]
}, Open  ]]
},
WindowSize->{1282, 1022},
WindowMargins->{{10, Automatic}, {Automatic, 9}},
PrintingCopies->1,
PrintingPageRange->{1, Automatic},
PrivateNotebookOptions->{"VersionedStylesheet"->{"Default.nb"[8.] -> False}},
ShowSelection->True,
FrontEndVersion->"9.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (January 25, \
2013)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3585586635-4340400"->{
  Cell[11466, 295, 463, 8, 109, "Print",
   CellTags->"Info3585586635-4340400"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3585586635-4340400", 25587, 712}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[557, 20, 293, 4, 29, "Subsubtitle"],
Cell[CellGroupData[{
Cell[875, 28, 401, 5, 92, "Title"],
Cell[1279, 35, 357, 6, 29, "Subsubtitle"],
Cell[1639, 43, 221, 6, 74, "Program"],
Cell[CellGroupData[{
Cell[1885, 53, 98, 1, 35, "Subsubsection"],
Cell[1986, 56, 872, 18, 228, "Program"]
}, Closed]],
Cell[CellGroupData[{
Cell[2895, 79, 156, 2, 36, "Subsection"],
Cell[3054, 83, 270, 7, 31, "Text"],
Cell[3327, 92, 635, 18, 63, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3999, 115, 181, 4, 44, "Subsection"],
Cell[4183, 121, 2080, 46, 46, "Input"],
Cell[6266, 169, 997, 17, 28, "Input"],
Cell[7266, 188, 2362, 53, 28, "Input"],
Cell[9631, 243, 178, 4, 30, "Text"],
Cell[9812, 249, 260, 6, 28, "Input"],
Cell[10075, 257, 407, 7, 49, "Text"],
Cell[10485, 266, 123, 2, 28, "Input"],
Cell[10611, 270, 379, 8, 28, "Input"],
Cell[10993, 280, 243, 6, 31, "Text"],
Cell[CellGroupData[{
Cell[11261, 290, 202, 3, 28, "Input"],
Cell[11466, 295, 463, 8, 109, "Print",
 CellTags->"Info3585586635-4340400"]
}, Open  ]],
Cell[11944, 306, 370, 10, 28, "Input"],
Cell[12317, 318, 239, 4, 30, "Text"],
Cell[12559, 324, 373, 10, 28, "Input"],
Cell[12935, 336, 176, 4, 28, "Input"],
Cell[13114, 342, 367, 7, 30, "Text"],
Cell[13484, 351, 120, 2, 28, "Input"],
Cell[13607, 355, 412, 9, 28, "Input"],
Cell[14022, 366, 431, 9, 28, "Input"],
Cell[14456, 377, 1422, 42, 28, "Input"],
Cell[15881, 421, 478, 11, 28, "Input"],
Cell[16362, 434, 225, 5, 30, "Text"],
Cell[16590, 441, 1270, 35, 28, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[17897, 481, 216, 3, 44, "Subsection"],
Cell[18116, 486, 328, 10, 28, "Input"],
Cell[18447, 498, 1318, 39, 28, "Input"],
Cell[19768, 539, 503, 10, 28, "Input"],
Cell[20274, 551, 186, 4, 30, "Text"],
Cell[20463, 557, 416, 9, 28, "Input"],
Cell[20882, 568, 380, 8, 28, "Input"],
Cell[21265, 578, 1376, 42, 28, "Input"],
Cell[22644, 622, 430, 11, 28, "Input"],
Cell[23077, 635, 315, 6, 30, "Text"],
Cell[23395, 643, 1190, 34, 28, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24622, 682, 97, 1, 44, "Subsection"],
Cell[24722, 685, 363, 8, 30, "Text"]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
